#!/usr/bin/env python  # pylint: disable=invalid-name,too-many-lines
# -*- encoding: utf-8 -*-

'''
A collection of code for manipulating SVG images. This can be used to extract,
embed and deduplicate images, remove superfluous elements, generate PDF
documents for JessyInk presentations, and much more.

The code was hacked together gradually to suit varoius needs and will be
refactored and split into separate modules in the future.
'''

from collections import OrderedDict, deque
from html.parser import HTMLParser
from xml.sax.saxutils import quoteattr
import argparse
import base64
import copy
import filecmp
import hashlib
import io
import itertools
import logging
import math
import mimetypes
import os.path
import re
import shutil
import sys
import time
import urllib.parse
import xml.etree.ElementTree as ET


# --------------------------------- Globals ---------------------------------- #
# When multiple extensions are possible for a given MIME-type (jpg, jpe, jpeg),
# the order in which mimetypes returns the preferred extension is random due
# to the use of a dictionary. Use this to set a preference.
PREFERRED_EXTS = (
    '.jpg',
)

BASE64EXTENSION = ';base64'
DEFAULT_MIMETYPE = 'text/plain'
ENCODING = 'utf-8'
FILE_SCHEMES = ('', 'file')
DATA_SCHEME = 'data'
DEFAULT_HASHTYPE = 'sha512'
BACKUP_EXT = '.bak'
TIMESTAMP_FMT = '.%Y%m%d.%H%M%S'

COMMON_NAMESPACES = {
    'cc': "http://creativecommons.org/ns#",
    'dc': "http://purl.org/dc/elements/1.1/",
    'inkscape': "http://www.inkscape.org/namespaces/inkscape",
    "jessyink": "https://launchpad.net/jessyink",
    'rdf': "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    'sodipodi': "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd",
    'svg': "http://www.w3.org/2000/svg",
    'xlink': "http://www.w3.org/1999/xlink",
}

JI_EFFECT_EFFECT_IN = 'effectIn'
JI_EFFECT_EFFECT_OUT = 'effectOut'
JI_EFFECT_VIEW = 'view'
JI_EFFECTS = (
    JI_EFFECT_EFFECT_IN,
    JI_EFFECT_EFFECT_OUT,
    JI_EFFECT_VIEW
)

JI_AUTOTEXT_SLIDE_TITLE = 'slideTitle'
JI_AUTOTEXT_SLIDE_NUMBER = 'slideNumber'
JI_AUTOTEXT_NUMBER_OF_SLIDES = 'numberOfSlides'
JI_AUTOTEXTS = (
    JI_AUTOTEXT_SLIDE_TITLE,
    JI_AUTOTEXT_SLIDE_NUMBER,
    JI_AUTOTEXT_NUMBER_OF_SLIDES
)

# Criteria for selecting autotext elements.
JI_AUTOTEXT_BY = {
    'id': 'element id attribute',
    'text': 'content of text element'
}
JI_AUTOTEXT_BY_DEFAULT = 'text'


LOGGER = logging.getLogger('SvgImgMgr')


# ---------------------------- Generic Functions ----------------------------- #

def hash_data(data, hashtype=DEFAULT_HASHTYPE):
    '''
    Hash the given data and return the hex digest.
    '''
    hsh = hashlib.new(hashtype)
    hsh.update(data)
    return hsh.hexdigest()


def hash_file(path, hashtype=DEFAULT_HASHTYPE):
    '''
    Hash a file.
    '''
    hsh = hashlib.new(hashtype)
    with open(path, 'rb') as hdl:
        # f._CHUNK_SIZE
        buf = hdl.read(io.DEFAULT_BUFFER_SIZE)
        while buf:
            hsh.update(buf)
            buf = hdl.read(io.DEFAULT_BUFFER_SIZE)
    return hsh.hexdigest()


def prompt_list(question, items):
    '''
    Prompt the user to select 1 item from a list. Return the item and its index.
    '''
    if not isinstance(items, list):
        items = list(items)

    if not items:
        return None, None
    if not items[1:]:
        return items[0], 0

    width = len(str(len(items)))
    fmt = f'[{{:{width:d}d}}] {{}}'
    lst = '\n'.join(fmt.format(i, item) for (i, item) in enumerate(items))
    while True:
        try:
            choice = int(input(f'{question}\n{lst}\n'))
            return items[choice], choice
        except (ValueError, IndexError):
            pass


def copy_file(srcpath, dstpath):
    '''
    Copy the source path to the destination path,
    '''
    shutil.copy(srcpath, dstpath)
    LOGGER.info('copied %s to %s', srcpath, dstpath)


def move_file(srcpath, dstpath):
    '''
    Move the source path to the destination path,
    '''
    shutil.move(srcpath, dstpath)
    LOGGER.info('moved %s to %s', srcpath, dstpath)


def indexed_paths(path, width=3, start=1, orig_first=False):
    '''
    Generator for indexed paths. The index is inserted as a sub-extenstion before
    the final extension.

    width:
      The number of digits to use in the inserted index, e.g. 3 -> 001, 002, etc.

    start:
      The start index.

    orig_first:
      If True, yield the original, unindexed path first.
    '''
    if orig_first:
        yield path
    base_path, ext = os.path.splitext(path)
    fmt = f'{{}}.{{:0{width:d}d}}{{}}'
    i = start
    while True:
        yield fmt.format(base_path, i, ext)
        i += 1


def first_free_path_unless_same(path, refpath=None, **kwargs):
    '''
    Find either a free path based on the input path, or a non-free path with the
    same content as the optional reference path. Free paths are constructed with
    indexed_paths().

    Return the found path and a boolean to indicate if the path is occupied with
    the same content as the reference path.

    The purpose of this function is to avoid duplicates when collecting linked
    images in a directory. If the target path contains the same data as the
    reference path then it can be used in the image without recopying the content.

    If refpath is None, this simply returns the first free path.

    Keyword arguments are passed through to indexed_paths().
    '''
    if not os.path.isfile(refpath):
        LOGGER.error('first_free_path_unless_same(): %s is not a file path, ignoring', refpath)
        refpath = None

    if refpath:
        for pth in indexed_paths(path, orig_first=True, **kwargs):
            try:
                if filecmp.cmp(pth, refpath, shallow=False):
                    return pth, True
            except FileNotFoundError:
                return pth, False

    else:
        for pth in indexed_paths(path, orig_first=True, **kwargs):
            if not os.path.exists(pth):
                return pth, False

    return None, False


# ----------------------- SVG Transformation Functions ----------------------- #
# https://developer.mozilla.org/en/docs/Web/SVG/Attribute/transform
# http://tutorials.jenkov.com/svg/svg-viewport-view-box.html
# https://sarasoueidan.com/blog/svg-coordinate-systems/
# Create our own definitions to avoid dependency on numpy or other external
# libraries.
def matrix_multiply(a, b):  # pylint: disable=invalid-name
    '''
    Multiply SVG transformation matrices. The matrices should be given as
    6-element items.
    '''
    # pylint: disable=invalid-name
    a1, a2, a3, a4, a5, a6 = a
    b1, b2, b3, b4, b5, b6 = b
    return (
        a1 * b1 + a3 * b2,
        a2 * b1 + a4 * b2,
        a1 * b3 + a3 * b4,
        a2 * b3 + a4 * b4,
        a1 * b5 + a3 * b6 + a5,
        a2 * b5 + a4 * b6 + a6
    )


def matrix_invert(m):  # pylint: disable=invalid-name
    '''
    Invert the matrix.
    '''
    # pylint: disable=invalid-name
    a, b, c, d, e, f = m
    D = a * d - b * c
    return (
        d / D,
        -b / D,
        -c / D,
        a / D,
        (c * f - d * e) / D,
        (b * e - a * f) / D
    )


def matrix_multiply_many(ms):  # pylint: disable=invalid-name
    '''
    Multiply a list of matrices.
    '''
    # pylint: disable=invalid-name
    result = (1., 0., 0., 1., 0., 0.)
    for m in ms[::-1]:
        result = matrix_multiply(m, result)
    return result


def matrix_translate(x, y=0.):  # pylint: disable=invalid-name
    '''
    Get a translation matrix.
    '''
    return (1., 0., 0., 1., x, y)


def matrix_scale(x, y=None):  # pylint: disable=invalid-name
    '''
    Get a scale matrix.
    '''
    if y is None:
        y = x
    return (x, 0., 0., y, 0., 0.)


def matrix_rotate(theta, x=0., y=0.):  # pylint: disable=invalid-name
    '''
    Get a rotation matrix about a point.
    '''
    # pylint: disable=invalid-name
    theta *= math.pi / 180.
    c = math.cos(theta)
    s = math.sin(theta)
    r = (c, s, -s, c, 0., 0.)
    if x != 0. or y != 0.:
        # rotate(a, x, y) = translate (x, y) rotate(a) translate(-x, -y)
        t_minus = matrix_translate(-x, -y)
        t_plus = matrix_translate(x, y)
        return matrix_multiply_many((t_plus, r, t_minus))
    return r


def matrix_skewX(a):  # pylint: disable=invalid-name
    '''
    Get a skew-x matrix.
    '''
    return (1., 0., math.tan(a), 1., 0., 0.)


def matrix_skewY(a):  # pylint: disable=invalid-name
    '''
    Get a skew-y matrix.
    '''
    return (1., math.tan(a), 0., 1., 0., 0.)


# ----------------------------- GenericException ----------------------------- #
class GenericException(Exception):
    '''
    Generic exceptions raised by this module.
    '''


# ------------------------------- XmlnsParser -------------------------------- #
class XmlnsParser(HTMLParser):  # pylint: disable=abstract-method
    '''
    Determine the XML namespaces in a document and the default. After feeding the
    document, the following attributes will be set:

    xmlns:
      Dictionary mapping namespaces ("xmlns:<ns>") to URIs.

    default:
      The URI of the default namespace ("xmlns" tag).
    '''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.xmlns = dict(COMMON_NAMESPACES.items())
        self.default = None

    def handle_starttag(self, tag, attrs):
        for attr, val in attrs:
            if attr.startswith('xmlns:'):
                namespace = attr[6:]
                self.xmlns[namespace] = val
            elif attr == 'xmlns':
                self.default = val


# ----------------------------------- NSET ----------------------------------- #
class NSET():
    '''
    Simple wrapper around ElementTree that parses and handles namespaces.
    '''

    def __init__(self):
        self.xmlns_default = None
        self.xmlns = {}
        self.rev_xmlns = {}
        self.xml = None
        self.root = None

    def __copy__(self):
        new = self.__class__()
        new.xmlns_default = self.xmlns_default
        new.xmlns = copy.copy(self.xmlns)
        new.rev_xmlns = dict((v, k) for k, v in new.xmlns.items())
        new.xml = copy.copy(self.xml)
        new.root = new.xml.getroot()

    def __deepcopy__(self, *args, **kwargs):
        new = self.__class__()
        new.xmlns_default = self.xmlns_default
        new.xmlns = copy.deepcopy(self.xmlns, *args, **kwargs)
        new.rev_xmlns = dict((v, k) for k, v in new.xmlns.items())
        new.xml = copy.deepcopy(self.xml, *args, **kwargs)
        new.root = new.xml.getroot()
        return new

    def load(self, path):
        '''
        Load the document from a path. Parse the namespaces.
        '''
        with open(path, 'r', encoding='utf-8') as hdl:
            xml = hdl.read()
        xmlns_parser = XmlnsParser()
        xmlns_parser.feed(xml)
        self.xmlns = xmlns_parser.xmlns
        self.xmlns_default = xmlns_parser.default
        self.rev_xmlns = dict((v, k) for k, v in self.xmlns.items())

        for namespace, uri in self.xmlns.items():
            if uri == self.xmlns_default:
                namespace = ''
            try:
                ET.register_namespace(namespace, uri)
            except ValueError:
                LOGGER.debug('skipping registration of namespace tag "%s"', namespace)
        self.xml = ET.parse(path)
        self.root = self.xml.getroot()
        LOGGER.debug('loaded %s', path)

    def save(self, path):
        '''
        Save the document to a path.
        '''
        path = os.path.abspath(path)
        timestamp = time.strftime(TIMESTAMP_FMT, time.localtime())
        try:
            copy_file(path, f'{path}{timestamp}{BACKUP_EXT}')
        except FileNotFoundError:
            parent = os.path.dirname(path)
            if parent:
                os.makedirs(parent, exist_ok=True)
        self.xml.write(
            path,
            xml_declaration=True,
            encoding=ENCODING,
            method='xml'
        )
        LOGGER.info('saved %s', path)

    def prefix_ns(self, ns, tag):  # pylint: disable=invalid-name
        '''
        Add the namespace prefix. This will look up the namespace and prefix
        "{<url>}" to the given tag/attribute.
        '''
        if ns is None:
            ns = self.xmlns_default
        elif not ns:
            return tag
        else:
            ns = self.xmlns[ns]
        if ns is None:
            return tag
        return f'{{{ns}}}{tag}'

    def display_tag(self, tag):
        '''
        Add the namespace prefix. This will look up the namespace and prefix
        "{<url>}" to the given tag/attribute.
        '''
        if tag[0] != '{':
            return tag
        # pylint: disable=invalid-name
        ns, tag = tag[1:].split('}', 1)
        if ns == self.xmlns_default:
            return tag
        try:
            return f'{self.rev_xmlns[ns]}:{tag}'
        except KeyError:
            return tag

    def iter(self, tag, ns=None):  # pylint: disable=invalid-name
        '''
        Find all tags with the given name. If no namespace is given, the default is
        used.
        '''
        if tag is None:
            yield from self.root.iter()
        else:
            tag = self.prefix_ns(ns, tag)
            yield from self.root.iter(tag)

    def format_xpath_attribute(self, attr, attr_ns=None, val=None):
        '''
        Format attributes for XPATH queries.
        '''
        attr = self.prefix_ns(attr_ns, attr)
        val = '' if val is None else f'={quoteattr(val)}'
        return f'@{attr}{val}'

    def find_by_attributes(
        self,
        attrs,
        tag=None,
        tag_ns=None,
        choose='and'
    ):
        '''
        Find elements by attributes.

        attrs:
          An iterable of 3-tuples. Each tuple contains the following:

              * attribute name or None
              * attribute namespace or None
              * attribute value or None

        '''
        attrs = f' {choose} '.join(
            self.format_xpath_attribute(attr, attr_ns=attr_ns, val=val)
            for attr, attr_ns, val in attrs
        )
        if tag is None:
            tag = '*'
            if tag_ns is not None:
                tag = self.prefix_ns(tag_ns, tag)
        else:
            tag = self.prefix_ns(tag_ns, tag)
        xpath = f'.//{tag}[{attrs}]'
        yield from self.root.findall(xpath)

    def find_by_attribute(
        self,
        attr,
        attr_ns=None,
        val=None,
        tag=None,
        tag_ns=None
    ):
        '''
        Find elements by attribute. This is a thin wrapper around find_by_attributes.
        '''
        return self.find_by_attributes(
            (
                (attr, attr_ns, val),
            ),
            tag=tag,
            tag_ns=tag_ns
        )

    def update_hrefs(self, element, svgpath, targetpath, relative=False):
        '''
        Update the following hrefs:

        * xlink:href
        * sodipodi:absref
        '''
        dirpath = os.path.dirname(svgpath)

        if relative:
            relpath = targetpath
            targetpath = os.path.join(dirpath, relpath)
        else:
            relpath = os.path.relpath(targetpath, start=dirpath)

        attrs = element.attrib
        href_attr = self.prefix_ns('xlink', 'href')
        attrs[href_attr] = relpath

        try:
            absref_attr = self.prefix_ns('sodipodi', 'absref')
        except KeyError:
            pass
        else:
            # Do not create if absent.
            if absref_attr in attrs:
                attrs[absref_attr] = os.path.realpath(targetpath)

    def remove_by_id(self, elem):
        '''
        Remove an element by its "id" attribute.
        '''
        eid = elem.attrib['id']
        for parent in self.root.findall(f".//*[@id='{eid}'].."):
            parent.remove(elem)
            LOGGER.debug('removed #%s', eid)

    def map_parents(self, elem=None, attrib=None):
        '''
        Iterate over child-parent pairs.

        elem:
          If given, returned values will be limited to descendents of this element,
          otherwise the root element.

        attrib:
          If given, then this attribute will be returned for each element, if
          available, otherwise the elements themselves.
        '''
        if elem is None:
            elem = self.root
        for parent in elem.iter():
            for child in parent:
                if attrib:
                    try:
                        yield child.attrib[attrib], parent.attrib[attrib]
                    except KeyError:
                        continue
                else:
                    yield child, parent

    def references(self, elem):
        '''
        Return the set of IDs referenced by the given element.
        '''
        href_attr = self.prefix_ns('xlink', 'href')
        id_url_regex = re.compile(r'^url\(#([^)]+)\)$')
        # xlink:ref attributes
        try:
            href = elem.attrib[href_attr]
        except KeyError:
            pass
        else:
            if href[0] == '#':
                yield href[1:]
        # Generic catcher for other attributes with URL values, e.g.
        # clip-path="url(#clipPath1234)"
        for value in elem.attrib.values():
            match = id_url_regex.match(value)
            if match:
                yield match.group(1)

        # url(#id) in style attribute
        try:
            style_string = elem.attrib['style']
        except KeyError:
            pass
        else:
            style = parse_style(style_string)
            for value in style.values():
                match = id_url_regex.match(value)
                if match:
                    yield match.group(1)

    # TODO
    # Maybe cache ID maps in instance attribute.

    def collect_transformations(self, elem):
        '''
        Collect all transformation in the element hierarchy above the given element.
        '''
        id_map = {e.attrib['id']: e for e in self.find_by_attribute('id', attr_ns='')}
        id_map[self.root.attrib['id']] = self.root
        parent_id_map = dict(self.map_parents(attrib='id'))
        matrix = parse_transform(elem)
        while True:
            try:
                eid = elem.attrib['id']
                pid = parent_id_map[eid]
            except KeyError:
                break
            else:
                elem = id_map[pid]
                mat = parse_transform(elem)
                matrix = matrix_multiply(mat, matrix)
        return matrix

    # TODO
    # Maybe cache ID maps in instance attribute.

    def visible_elements(self, elems=None):
        '''
        Return iterators over (directly or indirectly) visible elements and
        invisible elements.
        '''
        id_map = {e.attrib['id']: e for e in self.find_by_attribute('id', attr_ns='')}
        id_map[self.root.attrib['id']] = self.root
        parent_id_map = dict(self.map_parents(attrib='id'))
        if elems is None:
            elems = self.root.iter()
        elem_queue = deque(e for e in elems if not is_hidden(e))
        vids = set()
        while elem_queue:
            elem = elem_queue.popleft()
            try:
                eid = elem.attrib['id']
            except KeyError:
                eid = None
            else:
                vids.add(eid)
            # Get parent.
            try:
                pid = parent_id_map[eid]
                if pid not in vids:
                    parent_elem = id_map[pid]
                    elem_queue.append(parent_elem)
            except KeyError:
                pass
            # Get all referenced elements.
            for ref_id in self.references(elem):
                if ref_id in vids:
                    continue
                try:
                    ref_elem = id_map[ref_id]
                except KeyError:
                    if eid:
                        LOGGER.warning('undefined reference in element #%s: #%s', eid, ref_id)
                    else:
                        LOGGER.warning('undefined reference: %s', ref_id)
                else:
                    elem_queue.extend(ref_elem.iter())
        return vids

    def remove_invisible(self, elems=None):
        '''
        Simplify the document by removing invisible elements. Elements are
        considered invisible if they are hidden, not referenced by another visible
        element, and all of their nested elements are invisible.
        '''
        vids = self.visible_elements(elems)
        queue = deque((self.root,))
        removed = False
        while queue:
            elem = queue.popleft()
            try:
                eid = elem.attrib['id']
            except KeyError:
                queue.extend(elem)
            else:
                if eid in vids:
                    queue.extend(elem)
                else:
                    self.remove_by_id(elem)
                    removed = True
        return removed

    def removed_unused_defs(self):
        '''
        Remove unreferenced definitions from the document.
        '''
        used_ref_ids = set()
        for elem in self.root.iter():
            used_ref_ids.update(self.references(elem))
        removed = False
        tag = self.prefix_ns('svg', 'defs')
        for defs in self.root.findall(f'./{tag}'):
            for def_ in list(defs):
                try:
                    def_id = def_.attrib['id']
                except KeyError:
                    continue
                else:
                    if def_id not in used_ref_ids:
                        defs.remove(def_)
                        LOGGER.info('removed unused definition #%s', def_id)
                        removed = True
        return removed


# ------------------------------- JessyInkNSET ------------------------------- #

class JessyInkNSET(NSET):
    '''
    Subclass of NSET for working with JessyInk presentations.
    '''

    # TODO
    # Add options to return layers by name or number. For names, use
    # `label_attr = self.prefix_ns('inkscape', 'label')`.
    def layers(self):
        '''
        Yield all layers.
        '''
        yield from self.find_by_attribute(
            'groupmode',
            attr_ns='inkscape',
            val='layer'
        )

    def effects(self, per_effect=False):
        '''
        Yield the number of effects per layer, in order.
        '''
        master_attr = self.prefix_ns('jessyink', 'masterSlide')
        ji_effect_attrs = tuple(
            self.prefix_ns('jessyink', n) for n in JI_EFFECTS
        )
        for layer in self.layers():  # pylint: disable=too-many-nested-blocks
            if master_attr in layer.attrib:
                continue
            orders = {} if per_effect else set()
            # jessyink:effectIn="name:appear;order:1;length:800"
            # jessyink:effectOut="name:appear;order:1;length:800"
            # jessyink:view="name:view;order:4;length:300"
            for child in layer.iter():
                for name, attr in zip(JI_EFFECTS, ji_effect_attrs):
                    try:
                        value = child.attrib[attr]
                    except KeyError:
                        continue
                    effect = parse_style(value)
                    try:
                        ordr = int(effect['order'])
                    except KeyError:
                        continue
                    else:
                        if per_effect:
                            # pylint: disable=unsubscriptable-object,unsupported-assignment-operation
                            try:
                                orders[name].add(ordr)
                            except KeyError:
                                orders[name] = set((ordr,))
                        else:
                            orders.add(ordr)
            yield layer, orders

    def display_by_effect(self, elem, order):
        '''
        Set the visibility of an element based on the JessyInk effects and the given
        order.
        '''
        effectIn_attr = self.prefix_ns('jessyink', JI_EFFECT_EFFECT_IN)
        effectOut_attr = self.prefix_ns('jessyink', JI_EFFECT_EFFECT_OUT)
        view_attr = self.prefix_ns('jessyink', JI_EFFECT_VIEW)

        # View boxes are always invisible.
        try:
            view = int(parse_style(elem.attrib[view_attr])['order'])
        except KeyError:
            pass
        else:
            if order == view:
                return JI_EFFECT_VIEW
            set_style(elem, 'style', 'display', 'none')
            return None

        # Objects are only visible between effectIn and effectOut.
        try:
            orderIn = int(parse_style(elem.attrib[effectIn_attr])['order'])
        except KeyError:
            orderIn = 0
        try:
            orderOut = int(parse_style(elem.attrib[effectOut_attr])['order'])
        except KeyError:
            orderOut = None
        if order < orderIn \
                or (orderOut is not None and order >= orderOut):
            set_style(elem, 'style', 'display', 'none')
        else:
            set_style(elem, 'style', 'display', 'inherit')

        return None

    def clear_effects(self, layers=None, nested=False):  # pylint: disable=too-many-branches
        '''
        Clear effects. The master slide, if present, is uneffected.

        layers:
          The numbers of layers on which to clear effects. If None, all layers will
          be cleared of effects.

        nested:
          If True, all nested elements of an element with an effect will have their
          effects cleared. The parent element's effects will not be cleared.
        '''
        master_attr = self.prefix_ns('jessyink', 'masterSlide')
        ji_effect_attrs = tuple(
            self.prefix_ns('jessyink', n) for n in JI_EFFECTS
        )
        if layers is not None:
            layers = set(layers)
        for layer in self.layers():  # pylint: disable=too-many-nested-blocks
            if (layers and layer not in layers) \
                    or master_attr in layer.attrib:
                continue

            if nested:
                for child in layer.iter():
                    for attr in ji_effect_attrs:
                        if attr in child.attrib:
                            for grandchild in child.iter():
                                # Watch out for inbreds! :P
                                if grandchild == child:
                                    continue
                                for a in ji_effect_attrs:
                                    try:
                                        del grandchild.attrib[a]
                                    except KeyError:
                                        continue
                            break

            else:
                for child in layer.iter():
                    for attr in ji_effect_attrs:
                        try:
                            del child.attrib[attr]
                        except KeyError:
                            continue


# ----------------------------- Argument Parsing ----------------------------- #

def get_parser():
    '''
    Get the argument parser.
    '''
    parser = argparse.ArgumentParser(
        description='Manage images within an SVG file.'
    )
    subparsers = parser.add_subparsers(
        help='sub-command help',
        dest='command'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='Increase verbosity.'
    )
    parser.add_argument(
        '--debug', action='store_true',
        help='Enable debugging messages.'
    )

    # --------------------------------- list --------------------------------- #
    parser_list = subparsers.add_parser(
        'list',
        aliases=('ls',),
        help='list svg data'
    )
    parser_list.set_defaults(func=list_info)
    parser_list.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_list.add_argument(
        '-i', '--images', action='store_true',
        help='List images.'
    )
    parser_list.add_argument(
        '--ids', action='store_true',
        help='List id attributes (sometimes useful for debugging).'
    )
    parser_list.add_argument(
        '--hidden', action='store_true',
        help='List hidden elements.'
    )

    # ------------------------------- collect -------------------------------- #
    parser_extract = subparsers.add_parser(
        'extract',
        aliases=('ex',),
        description='collect linked and/or embedded images',
        help='Collect linked and/or embedded images.',
    )
    parser_extract.set_defaults(func=extract)
    parser_extract.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_extract.add_argument(
        'outdir',
        help='Collect files in this directory.'
    )
    parser_extract.add_argument(
        '-u', '--update', metavar='<path>',
        help='Generate a new SVG image with updated links here.'
    )
    parser_extract.add_argument(
        '-e', '--embedded', action='store_true',
        help='Extract embedded data.'
    )
    parser_extract.add_argument(
        '-l', '--linked', action='store_true',
        help='Extract linked data.'
    )

    # -------------------------------- embed --------------------------------- #
    parser_embed = subparsers.add_parser(
        'embed',
        aliases=('em',),
        description='embed linked data',
        help='Embed linked data (usually images) with base64-encoded data URIs.'
    )
    parser_embed.set_defaults(func=embed)
    parser_embed.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_embed.add_argument(
        'outpath',
        help='The path to the new SVG file.'
    )

    # -------------------------------- rename -------------------------------- #
    parser_rename = subparsers.add_parser(
        'rename',
        aliases=('mv',),
        description='rename linked data',
        help='rename linked data (usually images)'
    )
    parser_rename.set_defaults(func=rename)
    parser_rename.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_rename.add_argument(
        '-p', '--pair', nargs=2, action='append', metavar=('<oldpath>', '<newpath>'), default=[],
        help='Old and new path of file to rename.'
    )
    parser_rename.add_argument(
        '-c', '--copy', metavar='<path>',
        help='Copy files instead of moving them. The new SVG file will be saved to <path>.'
    )

    # ----------------------------- deduplicate ------------------------------ #
    parser_deduplicate = subparsers.add_parser(
        'dedupe',
        aliases=('dd',),
        description='deduplicate images',
        help='deduplicae linked and embedded images'
    )
    parser_deduplicate.set_defaults(func=deduplicate)
    parser_deduplicate.add_argument(
        'svg',
        help='The SVG file.'
    )

    # -------------------------------- scrub --------------------------------- #
    parser_scrub = subparsers.add_parser(
        'scrub',
        aliases=('sc',),
        description='remove unnecessary elements',
        help='remove unnecessary elements'
    )
    parser_scrub.set_defaults(func=scrub)
    parser_scrub.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_scrub.add_argument(
        '-o', '--output', metavar='<path>',
        help='Save the updated file here. Without this option, the original will be overwritten.'
    )
    parser_scrub.add_argument(
        '-l', '--list', action='store_true',
        help='List changes that would be made.'
    )
    parser_scrub.add_argument(
        '-a', '--all', action='store_true',
        help='Remove all unnecessary elements.'
    )
    parser_scrub.add_argument(
        '--sodipodi', action='store_true',
        help='Remove sodipodi:absref tags.'
    )
    parser_scrub.add_argument(
        '--defs', action='store_true',
        help='Remove unused definitions.'
    )

    # ------------------------------- inkscape ------------------------------- #
    parser_ink = subparsers.add_parser(
        'inkscape',
        aliases=('ink',),
        description='Inkscape commands',
        help='Inkscape commands'
    )
    ink_subparsers = parser_ink.add_subparsers(
        help='sub-command help',
        dest='ink_command'
    )

    # ------------------------- deduplicate markers -------------------------- #
    parser_dedupe_markers = ink_subparsers.add_parser(
        'dedupemarkers',
        aliases=('dm',),
        description='deduplicate Inkscape stock markers',
        help='deduplicate Inkscape stock markers'
    )
    parser_dedupe_markers.set_defaults(func=dedupe_markers)
    parser_dedupe_markers.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_dedupe_markers.add_argument(
        '-o', '--output', metavar='<path>',
        help='Save the updated file here. Without this option, the original will be overwritten.'
    )
    parser_dedupe_markers.add_argument(
        '-l', '--list', action='store_true',
        help='List changes that would be made.'
    )
    parser_dedupe_markers.add_argument(
        '-d', '--delete', action='store_true',
        help='Delete unused markers.'
    )

    # ------------------------------- jessyink ------------------------------- #
    parser_ji = subparsers.add_parser(
        'jessyink',
        aliases=('ji',),
        description='JessyInk commands',
        help='JessyInk commands'
    )
    ji_subparsers = parser_ji.add_subparsers(
        help='sub-command help',
        dest='ji_command'
    )

    # ---------------------------- jessyink list ----------------------------- #
    parser_ji_slides = ji_subparsers.add_parser(
        'list',
        aliases=('ls',),
        description='list JessyInk slides',
        help='list JessyInk slides'
    )
    parser_ji_slides.set_defaults(func=ji_list)
    parser_ji_slides.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_ji_slides.add_argument(
        '-c', '--count', action='store_true',
        help='Count the number of effects per slide.'
    )
    parser_ji_slides.add_argument(
        '-t', '--title', action='store_true',
        help='Include the slide title in the output.'
    )

    # -------------------------- jessyink renumber --------------------------- #
    parser_ji_renumber = ji_subparsers.add_parser(
        'renumber',
        aliases=('rn',),
        description='renumber JessyInk effects',
        help='renumber JessyInk effects to fill numbering gaps'
    )
    parser_ji_renumber.set_defaults(func=ji_renumber_effects)
    parser_ji_renumber.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_ji_renumber.add_argument(
        '-l', '--list', action='store_true',
        help='List changes that would be made.'
    )
    parser_ji_renumber.add_argument(
        '-o', '--output', metavar='<path>',
        help='Save the updated file here. Without this option, the original will be overwritten.'
    )

    # -------------------------- jessyink autotext --------------------------- #
    parser_ji_autotext = ji_subparsers.add_parser(
        'autotext',
        aliases=('at',),
        description='set and unset JessyInk autotexts',
        help='set and unset JessyInk autotexts'
    )
    parser_ji_autotext.set_defaults(func=ji_autotexts)
    parser_ji_autotext.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_ji_autotext.add_argument(
        'args', nargs='*',
        help='The elements to which to apply the changes.'
    )
    parser_ji_autotext.add_argument(
        '-l', '--list', action='store_true',
        help='List autotexts.'
    )
    parser_ji_autotext.add_argument(
        '-t', '--type', choices=JI_AUTOTEXTS,
        help=('''
            Select the type of autotext to set. If this option is not given,
            autotext will be unset on the selected targets, or all if no targets
            are given.
        ''')
    )
    parser_ji_autotext.add_argument(
        '--by', choices=JI_AUTOTEXT_BY, default=JI_AUTOTEXT_BY_DEFAULT,
        help=(
            'Criterion for setting autotexts. '
            f"{', '.join(f'{k}: {v}' for k, v in sorted(JI_AUTOTEXT_BY.items()))} "
            'Default: %(default)s'
        )
    )
    parser_ji_autotext.add_argument(
        '-r', '--remove', action='store_true',
        help='Remove autotexts on the selected elements, or all if no arguments are given.'
    )
    parser_ji_autotext.add_argument(
        '-o', '--output', metavar='<path>',
        help='Save the updated file here. Without this option, the original will be overwritten.'
    )

    # --------------------------- jessyink export ---------------------------- #
    parser_ji_export = ji_subparsers.add_parser(
        'export',
        aliases=('ex',),
        description='export JessyInk slides',
        help='export JessInk slides'
    )
    parser_ji_export.set_defaults(func=ji_export)
    parser_ji_export.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_ji_export.add_argument(
        'outdir',
        help='The output directory.'
    )
    parser_ji_export.add_argument(
        '-s', '--slides', nargs='+',
        help=('''
            Select slides to export, For example export the first 5 slides with
            all effects with "-s 1 2 3 4 5", or in Bash use "-s {1..5}". To
            select a slide and an effect, use "/". For example, the first 3
            effects of the 8th slide: "-s 8/0 8/1 8/2" or "-s 8/{0..2}" (effects
            are numbered from 0). Negative numbers will be counted from the end.
            To export just the final effect of each slide, use "-s /-1".'
        ''')
    )
    parser_ji_export.add_argument(
        '--simplify', action='store_true',
        help='Remove unnecessary elements from each slide to reduce its size.'
    )

    # --------------------------- jessyink export ---------------------------- #
    parser_ji_clear = ji_subparsers.add_parser(
        'clear',
        aliases=('cl',),
        description='clear JessyInk effects',
        help='clear JessyInk effects'
    )
    parser_ji_clear.set_defaults(func=ji_clear)
    parser_ji_clear.add_argument(
        'svg',
        help='The SVG file.'
    )
    parser_ji_clear.add_argument(
        '--layers', type=int, nargs='+', metavar='<number>',
        help='The layers on which to clear effects.'
    )
    parser_ji_clear.add_argument(
        '--nested', action='store_true',
        help='Only clear nested effects.'
    )
    parser_ji_clear.add_argument(
        '-o', '--output', metavar='<path>',
        help='Save the updated file here. Without this option, the original will be overwritten.'
    )

    return parser


# ---------------------------- Accessor Functions ---------------------------- #

def int_formatter(number):
    '''
    Get a format string for all ints up to n.
    '''
    return f'{{:0{len(str(number)):d}d}}'


def parse_style(string):
    '''
    Parse string of key-value pairs: "key1:value1;key2:value2"
    '''
    string = string.strip(' \t\r\n;')
    style = OrderedDict()
    if not string:
        return style
    for pair in string.split(';'):
        try:
            key, value = pair.split(':', 1)
            key = key.strip()
            if not key:
                raise ValueError('missing key')
        except ValueError:
            LOGGER.warning('ignoring invalid style setting: %s', pair)
            continue
        style[key.strip()] = value
    return style


def unparse_style(odict):
    '''
    Reconstruct the style attribute from key-value pairs: "key1:value1;key2:value2"
    '''
    try:
        pairs = odict.items()
    except AttributeError:
        pairs = odict
    return ';'.join(f'{k}:{v}' for k, v in pairs)


def get_style(element, attr, key):
    '''
    Set a value in a style-like attribute.
    '''
    try:
        string = element.attrib[attr]
    except KeyError:
        return None
    odict = parse_style(string)
    return odict.get(key, None)


def set_style(elem, attr, key, value):
    '''
    Set a value in a style-like attribute.
    '''
    try:
        string = elem.attrib[attr]
    except KeyError:
        elem.attrib[attr] = f'{key}:{value}'
    else:
        odict = parse_style(string)
        odict[key] = value
        elem.attrib[attr] = unparse_style(odict)


def parse_transform(elem):
    '''
    Return a 6-tuple "matrix" for the given element by parsing the `transform`
    attribute, if available.
    '''
    # pylint: disable=invalid-name
    m = matrix_translate(0.)
    try:
        transform_str = elem.attrib['transform']
    except KeyError:
        return m
    function_regex = re.compile(r'(\S+)\(([^)]*)\)')
    for (func, args) in function_regex.findall(transform_str):
        args = tuple(float(x) for x in args.replace(',', ' ').split())
        if func == 'matrix':
            n = tuple(args)
        elif func == 'translate':
            n = matrix_translate(*args)
        elif func == 'scale':
            n = matrix_scale(*args)
        elif func == 'rotate':
            n = matrix_rotate(*args)
        elif func == 'skewX':
            n = matrix_skewX(*args)
        elif func == 'skewY':
            n = matrix_skewY(*args)
        else:
            try:
                LOGGER.warning(
                    'unrecognized transformation in element %s: %s',
                    elem.attrib['id'],
                    func
                )
            except KeyError:
                LOGGER.warning(
                    'unrecognized transformation in %s element: %s',
                    elem.tag,
                    func
                )
            continue
        m = matrix_multiply(m, n)
    return m


# --------------------------------- Commands --------------------------------- #

def get_ext(mime):
    '''
    Get an extension for a MIME-type.
    '''
    exts = mimetypes.guess_all_extensions(mime, strict=False)
    for ext in exts:
        if ext in PREFERRED_EXTS:
            return ext
    # Sort to make deterministic.
    return sorted(exts)[0] if exts else None


def extract_data(
    data,
    outdir=None,
    name=None,
    prefix='extracted_',
    no_clobber=True
):
    '''
    Extract embedded data to path.
    '''
    mime, data = data.split(',', 1)
    if mime.endswith(BASE64EXTENSION):
        mime = mime[:-len(BASE64EXTENSION)]
        data = base64.b64decode(data)
    if not mime:
        mime = DEFAULT_MIMETYPE

    if outdir is None:
        return data

    ext = get_ext(mime)
    if name is None:
        name = f'{prefix}{hash_data(data)}'
    name += ext
    outpath = os.path.join(outdir, name)

    if no_clobber and os.path.exists(outpath):
        LOGGER.info('refusing to clobber %s', outpath)
    else:
        os.makedirs(outdir, exist_ok=True)
        with open(outpath, 'wb') as hdl:
            hdl.write(data)
        LOGGER.info('extracted %s', outpath)
    return outpath


def embed_data(parsed_uri, mimetype=None):
    '''
    Embed linked data.
    '''
    if not parsed_uri.scheme:
        with open(parsed_uri.path, 'rb') as hdl:
            data = hdl.read()
    else:
        with urllib.request.urlopen(urllib.parse.urlunparse(parsed_uri)) as hdl:
            data = hdl.read()
    data = base64.b64encode(data).decode(ENCODING)
    if mimetype is None:
        for mimetype in mimetypes.guess_type(parsed_uri.path, strict=False):  \
                # pylint: disable=redefined-argument-from-local
            if mimetype is not None:
                break
    if mimetype is None:
        mimetype = DEFAULT_MIMETYPE
    return f'{mimetype};base64,{data}'


def list_info(pargs):
    '''
    List various information.
    '''
    nset = NSET()
    nset.load(pargs.svg)
    svgdir = os.path.dirname(pargs.svg)
    if pargs.images:
        href_attr = nset.prefix_ns('xlink', 'href')
        for image in nset.iter('image'):
            attrs = image.attrib
            uri = attrs[href_attr]
            parsed_uri = urllib.parse.urlparse(uri)
            if parsed_uri.scheme == DATA_SCHEME:
                print(attrs['id'], f"embedded:{parsed_uri.path.split(',', 1)[0]}")
            else:
                path = os.path.join(svgdir, parsed_uri.path)
                tag = '' if os.path.exists(path) else ' [not found]'
                print(attrs['id'], f'{uri}{tag}')
    if pargs.ids:
        for elem in nset.iter(None):
            try:
                print(elem.attrib['id'], nset.display_tag(elem.tag))
            except KeyError:
                LOGGER.warning('%s element without an id tag', elem.tag)
    if pargs.hidden:
        for elem in nset.iter(None):
            if is_hidden(elem):
                try:
                    print('hidden:', elem.attrib['id'], nset.display_tag(elem.tag))
                except KeyError:
                    LOGGER.warning('hidden %s element without an id tag', elem.tag)


def extract(pargs):
    '''
    Extract linked and embedded images to a directory.
    '''
    nset = NSET()
    nset.load(pargs.svg)

    if pargs.linked or pargs.embedded:
        svgdir = os.path.dirname(pargs.svg)
        outdir = pargs.outdir
        os.makedirs(outdir, exist_ok=True)

        new_svgdir = os.path.dirname(pargs.update) if pargs.update else None

        href_attr = nset.prefix_ns('xlink', 'href')
        for image in nset.iter('image'):
            attrs = image.attrib
            uri = attrs[href_attr]
            parsed_uri = urllib.parse.urlparse(uri)
            dstpath = None
            if parsed_uri.scheme == DATA_SCHEME and pargs.embedded:
                dstpath = extract_data(parsed_uri.path, outdir=outdir)
            elif parsed_uri.scheme in FILE_SCHEMES and pargs.linked:
                srcpath = os.path.join(svgdir, parsed_uri.path)
                dstpath = os.path.join(outdir, os.path.basename(parsed_uri.path))
                dstpath, dstpath_exists = first_free_path_unless_same(dstpath, refpath=srcpath)
                if not dstpath_exists:
                    try:
                        copy_file(srcpath, dstpath)
                        LOGGER.info('copied %s to %s', srcpath, dstpath)
                    except FileNotFoundError as err:
                        LOGGER.warning('%s', err)
            if dstpath is not None and new_svgdir is not None:
                nset.update_hrefs(image, pargs.update, dstpath)

    if pargs.update:
        nset.save(pargs.update)


def embed(pargs):
    '''
    Embed all images.
    '''
    nset = NSET()
    nset.load(pargs.svg)
    href_attr = nset.prefix_ns('xlink', 'href')
    for image in nset.iter('image'):
        attrs = image.attrib
        uri = attrs[href_attr]
        parsed_uri = urllib.parse.urlparse(uri)
        if parsed_uri.scheme != DATA_SCHEME:
            attrs[href_attr] = urllib.parse.urlunparse((
                DATA_SCHEME,
                '',
                embed_data(parsed_uri),
                '',
                '',
                ''
            ))
    nset.save(pargs.outpath)


def deduplicate(pargs):
    '''
    Deduplicate images by relinking linked duplicates to a single external image.
    Embedded images that are duplicates of an external image will be linked to it.
    If there are multiple external duplicates, the user will be prompted to
    select one.
    '''
    nset = NSET()
    svgpath = os.path.abspath(pargs.svg)
    nset.load(svgpath)
    svgdir = os.path.dirname(svgpath)

    # Map checksums of images to all images that share them to determine
    # duplicates.
    images = {}
    href_attr = nset.prefix_ns('xlink', 'href')

    for image in nset.iter('image'):
        attrs = image.attrib
        uri = attrs[href_attr]
        parsed_uri = urllib.parse.urlparse(uri)

        if parsed_uri.scheme == DATA_SCHEME:
            data = extract_data(parsed_uri.path)
            checksum = hash_data(data)
            item = (image, None)

        elif parsed_uri.scheme in FILE_SCHEMES:
            path = os.path.join(svgdir, parsed_uri.path)
            checksum = hash_file(path)
            item = (image, parsed_uri.path)

        try:
            images[checksum].append(item)
        except KeyError:
            images[checksum] = [item]

    for checksum, items in images.items():
        # If there are no linked images, leave the data embedded.
        # If there is only one linked image, link all embedded images to it.
        # If there are more than one identical linked images, prompt the user to
        # select one.
        linked_items = list(x for x in items if x[1] is not None)
        linked_paths = sorted(set(x[1] for x in linked_items))
        n_linked = len(linked_paths)
        if n_linked == 0:
            continue
        if n_linked == 1:
            linked_path = linked_paths[0]
        elif n_linked > 1:
            linked_path, _ = prompt_list(
                'Select the linked image to use among these duplicates:',
                linked_paths
            )
            for image, old_path in linked_items:
                if old_path != linked_path:
                    nset.update_hrefs(image, svgpath, linked_path, relative=True)
                    LOGGER.info('relinked %s to %s', old_path, linked_path)

        for image, path in items:
            if path is None:
                nset.update_hrefs(image, svgpath, linked_path, relative=True)
                LOGGER.info('replaced embedded image with link to %s', linked_path)

    nset.save(svgpath)


def rename(pargs):
    '''
    Rename linked images.
    '''
    nset = NSET()
    nset.load(pargs.svg)
    svgdir = os.path.dirname(pargs.svg)
    outpath = pargs.copy if pargs.copy else pargs.svg

    pairs = tuple((
        os.path.realpath(oldpath),
        os.path.realpath(newpath)
    ) for oldpath, newpath in pargs.pair)

    abspaths = dict((
        os.path.abspath(oldpath),
        newpath
    ) for oldpath, newpath in pairs)

    href_attr = nset.prefix_ns('xlink', 'href')
    for image in nset.iter('image'):
        attrs = image.attrib
        uri = attrs[href_attr]
        parsed_uri = urllib.parse.urlparse(uri)
        if parsed_uri.scheme in FILE_SCHEMES:
            imgpath = os.path.join(svgdir, parsed_uri.path)
            abspath = os.path.abspath(imgpath)
            try:
                newpath = abspaths[abspath]
            except KeyError:
                pass
            else:
                nset.update_hrefs(image, outpath, newpath)
                break

    for oldpath, newpath in pairs:
        parent = os.path.dirname(newpath)
        if parent:
            os.makedirs(parent, exist_ok=True)
        try:
            if pargs.copy:
                copy_file(oldpath, newpath)
            else:
                move_file(oldpath, newpath)
        except FileNotFoundError as err:
            LOGGER.warning('%s', err)
    nset.save(outpath)


def scrub(pargs):
    '''
    Remove absref attributes.
    '''
    nset = NSET()
    nset.load(pargs.svg)

    absref_attr = nset.prefix_ns('sodipodi', 'absref')
    changed = False

    if pargs.all or pargs.sodipodi:
        for elem in nset.root.findall(f'.//*[@{absref_attr}]'):
            LOGGER.info(
                'scrubbing sodipodi absref from %s: %s',
                elem.attrib['id'],
                elem.attrib[absref_attr]
            )
            try:
                del elem.attrib[absref_attr]
            except KeyError:
                pass
            else:
                changed = True

    if pargs.all or pargs.defs:
        changed = nset.removed_unused_defs() or changed

    if not pargs.list and (changed or pargs.output != pargs.svg):
        nset.save(pargs.output if pargs.output else pargs.svg)


def fingerprint_marker(marker):
    '''
    Return a string identifier for a marker based on its style.
    '''
    path = marker.getchildren()[0]
    try:
        style_str = path.attrib['style']
    except KeyError:
        return ''
    style = parse_style(style_str)
    return unparse_style(sorted(style.items()))


def dedupe_markers(pargs):  # pylint: disable=too-many-branches
    '''
    Deduplicate Inkscape markers.
    '''
    nset = NSET()
    nset.load(pargs.svg)
    stockid_attr = nset.prefix_ns('inkscape', 'stockid')

    stock_markers = {}
    marker_elems = {}

    for marker in nset.find_by_attribute(
        'isstock',
        attr_ns='inkscape',
        val='true',
        tag='marker'
    ):
        # Append the style-based fingerprint to the stock ID to provide a unique
        # identifier for each colored marker.
        markerid = f'{marker.attrib[stockid_attr]}-{fingerprint_marker(marker)}'
        id_ = marker.attrib['id']
        marker_elems[id_] = marker
        try:
            # avoid dupes in list
            if id_ not in stock_markers[markerid]:
                stock_markers[markerid].append(id_)
        except KeyError:
            stock_markers[markerid] = [id_]

    replacements = {}
    for markerid, ids in list(stock_markers.items()):
        if len(ids) < 2:
            del stock_markers[markerid]
            continue
        stockid = markerid.split('-', 1)[0]
        if stockid in ids:
            replacements[markerid] = stockid
        else:
            replacements[markerid] = ids[0]

    if not replacements:
        return

    for markerid, dupes in sorted(stock_markers.items()):
        target = replacements[markerid]
        stock_markers[markerid] = [d for d in dupes if d != target]

    if pargs.list:
        for markerid, dupes in sorted(stock_markers.items()):
            target = replacements[markerid]
            stockid, fingerprint = markerid.split('-', 1)
            print('{} ({})\n  {}'.format(stockid, fingerprint, '\n  '.join(sorted(
                f'{d} -> {target}' for d in dupes
            ))))
        return

    for markerid, dupes in stock_markers.items():
        for dup in dupes:
            replacements[dup] = replacements[markerid]

    # XPATH wasn't working the way I expected and I didn't have time to debug.
    # Brute-force the search.
    for elem in nset.root.iter():
        try:
            style_str = elem.attrib['style']
        except KeyError:
            continue
        style = parse_style(style_str)
        at_least_one = False
        for key, val in list(style.items()):
            if val.startswith('url(#') and val.endswith(')'):
                url_id = val[5:-1]
                try:
                    replacement = replacements[url_id]
                except KeyError:
                    continue
                else:
                    if url_id == replacement:
                        continue
                    style[key] = f'url(#{replacement})'
                    at_least_one = True

        if at_least_one:
            style_str = unparse_style(style)
            elem.attrib['style'] = style_str

    if pargs.delete:
        for dupes in stock_markers.values():
            for dup in dupes:
                elem = marker_elems[dup]
                nset.remove_by_id(elem)
                LOGGER.info('removed marker #%s', id_)

    nset.save(pargs.output if pargs.output else pargs.svg)


def is_hidden(elem):
    '''
    Return True if the SVG element is hidden.
    '''
    try:
        return parse_style(elem.attrib['style'])['display'] == 'none'
    except KeyError:
        return False


def ji_list(pargs):
    '''
    List or count expected slides.
    '''
    nset = JessyInkNSET()
    nset.load(pargs.svg)
    title_attr = nset.prefix_ns('inkscape', 'label')
    for i, (layer, effects) in enumerate(nset.effects(), start=1):
        #     try:
        #       effect_number = len(effects[JI_EFFECT_EFFECT_IN])
        #     except KeyError:
        #       effect_number = 0
        #     try:
        #       effect_number = max(effect_number, len(effects[JI_EFFECT_EFFECT_OUT])+1)
        #     except KeyError:
        #       pass
        #     effect_number += 1
        effect_number = len(effects)
        if pargs.count:
            if pargs.title:
                print(i, effect_number, layer.attrib[title_attr])
            else:
                print(i, effect_number)
        else:
            #       if pargs.title:
            #         print(i, effects, layer.attrib[title_attr])
            #       else:
            #         print(i, effects)
            for j in range(effect_number + 1):
                if pargs.title:
                    print(i, j, layer.attrib[title_attr])
                else:
                    print(i, j)


def ji_clear(pargs):
    '''
    Clear effects from slides.
    '''
    nset = JessyInkNSET()
    nset.load(pargs.svg)
    nset.clear_effects(layers=pargs.layers, nested=pargs.nested)
    nset.save(pargs.output if pargs.output else pargs.svg)


def ji_renumber_effects(pargs):
    '''
    Renumber effects to fill meaningless gaps (e.g. 1,3,4 -> 1,2,3).
    '''
    nset = JessyInkNSET()
    nset.load(pargs.svg)
    attrs = tuple(nset.prefix_ns('jessyink', e) for e in JI_EFFECTS)
    for i, (layer, order) in enumerate(nset.effects(), start=1):
        # Determine if there are gaps in the ordering.
        order.add(0)
        order = sorted(order)
        if any((j != o) for j, o in enumerate(order)):
            if pargs.list:
                print(f"layer {i:d}: {' '.join(str(o) for o in order)}")
                continue
            order_map = dict((str(o), str(k)) for k, o in enumerate(order))
            for child in layer.iter():
                for attr in attrs:
                    try:
                        value = child.attrib[attr]
                    except KeyError:
                        continue
                    val_style = parse_style(value)
                    try:
                        val_style['order'] = order_map[val_style['order']]
                    except KeyError:
                        continue
                    else:
                        child.attrib[attr] = unparse_style(val_style)
    if not pargs.list:
        nset.save(pargs.output if pargs.output else pargs.svg)


# TODO: optimize
def ji_autotexts(pargs):  # pylint: disable=too-many-branches
    '''
    Set or unset JessyInk autotexts.
    '''
    nset = JessyInkNSET()
    nset.load(pargs.svg)
    autotext_attr = nset.prefix_ns('jessyink', 'autoText')

    if pargs.list or pargs.remove:
        elems = nset.find_by_attribute('autoText', attr_ns='jessyink', val=pargs.type)
        if pargs.args:
            if pargs.by == 'id':
                elems = (elem for elem in elems if elem in pargs.args)
            elif pargs.by == 'text':
                elems = (elem for elem in elems if elem.text in pargs.args)
        if pargs.remove:
            for elem in elems:
                del elem.attrib[autotext_attr]
        else:
            for elem in elems:
                print(f"id={quoteattr(elem.attrib['id'])} {elem.text}")
            return

    elif not pargs:
        LOGGER.error('no arguments given for selecting autotext elements')
    else:
        if pargs.by == 'id':
            attrs = (('id', None, a) for a in pargs.args)
            elems = nset.find_by_attributes(attrs)
        elif pargs.by == 'text':
            elems = (elem for elem in nset.root.iter() if elem.text in pargs.args)
        else:
            LOGGER.error('unhandled option passed to --by')
            return
        if not pargs.type:
            for elem in elems:
                del elem.attrib[autotext_attr]
        else:
            for elem in elems:
                elem.attrib[autotext_attr] = pargs.type

    nset.save(pargs.output if pargs.output else pargs.svg)


#
#   for i, (layer, order) in enumerate(nset.effects(), start=1):
#     # Determine if there are gaps in the ordering.
#     order.add(0)
#     order = sorted(order)
#     if any((j != o) for j, o in enumerate(order)):
#       if pargs.list:
#         print('layer {:d}: {}'.format(i, ' '.join(str(o) for o in order)))
#         continue
#       else:
#         order_map = dict((str(o), str(k)) for k, o in enumerate(order))
#         for child in layer.iter():
#           for attr in attrs:
#             try:
#               value = child.attrib[attr]
#             except KeyError:
#               continue
#             else:
#               vs = parse_style(value)
#               try:
#                 vs['order'] = order_map[vs['order']]
#               except KeyError:
#                 continue
#               else:
#                 child.attrib[attr] = unparse_style(vs)
#   if not pargs.list:
#     nset.save(pargs.output if pargs.output else pargs.svg)


def ji_export(pargs):  # pylint: disable=too-many-branches
    '''
    Export SVG images for each slide.
    '''
    nset = JessyInkNSET()
    svgpath = pargs.svg
    svgdir = os.path.dirname(svgpath)
    nset.load(svgpath)
    outdir = pargs.outdir
    os.makedirs(outdir, exist_ok=True)

    master_attr = nset.prefix_ns('jessyink', 'masterSlide')
    title_attr = nset.prefix_ns('inkscape', 'label')
    autotext_attr = nset.prefix_ns('jessyink', 'autoText')

#   nset.root.attrib['width'] += ', 100%'
#   nset.root.attrib['height'] += ', 100%'

    master_layer = None
    n_layers = 0
    for layer in nset.layers():
        if master_attr in layer.attrib:
            master_layer = layer
        else:
            n_layers += 1

    # Create a dictionary mapping selected slides to effects. If the slide number
    # is in the dictionary then it as been selected. If the effects are None then
    # all effects should be exported.
    selected_slides = {}
    if pargs.slides:
        for slide in pargs.slides:
            try:
                slide, effect = slide.split('/', 1)
            except ValueError:
                effect = None
            try:
                if slide:
                    slide = int(slide)
                    if slide < 0:
                        slide += 1 + n_layers
                else:
                    slide = None
                if effect is not None:
                    effect = int(effect)
            except ValueError as err:
                raise GenericException(f'invalid slide selection: {slide}') from err
            if effect is None:
                selected_slides[slide] = None
            else:
                try:
                    effects = selected_slides[slide]
                except KeyError:
                    selected_slides[slide] = set((effect,))
                else:
                    if effects is not None:
                        effects.add(effect)

    if master_layer:
        set_style(master_layer, 'style', 'display', 'inline')
    for script in nset.iter('script'):
        if script.attrib['id'] == 'JessyInk':
            nset.root.remove(script)

    href_attr = nset.prefix_ns('xlink', 'href')
    dummy_path = os.path.join(os.path.abspath(outdir), 'foo')
    # Update image paths for output directory.
    for image in nset.iter('image'):
        attrs = image.attrib
        uri = attrs[href_attr]
        parsed_uri = urllib.parse.urlparse(uri)
        if parsed_uri.scheme in FILE_SCHEMES:
            imgpath = os.path.abspath(os.path.join(svgdir, parsed_uri.path))
            nset.update_hrefs(image, dummy_path, imgpath)

    layer_fmt = int_formatter(n_layers)
    for i, (layer, order) in enumerate(nset.effects(), start=1):
        if selected_slides \
                and (i not in selected_slides and None not in selected_slides):
            continue

        order.add(0)
        n_order = len(order)
        effect_format = int_formatter(len(order))
        selected_effects = None
        try:
            selected_effects = selected_slides[i]
        except KeyError:
            try:
                selected_effects = selected_slides[None]
            except KeyError:
                pass
        if selected_effects is None:
            selected_effects = range(n_order)
        selected_effects = set(n_order + o if o < 0 else o for o in selected_effects)

        title = layer.attrib[title_attr]
        if master_layer:
            elements = itertools.chain(layer.iter(), master_layer.iter())
        else:
            elements = layer.iter()
        for elem in elements:
            try:
                autotext = elem.attrib[autotext_attr]
            except KeyError:
                pass
            else:
                if autotext == JI_AUTOTEXT_SLIDE_TITLE:
                    elem.text = title
                elif autotext == JI_AUTOTEXT_SLIDE_NUMBER:
                    elem.text = str(i)
                elif autotext == JI_AUTOTEXT_NUMBER_OF_SLIDES:
                    elem.text = str(n_layers)

        set_style(layer, 'style', 'display', 'inline')
        slide_nset = copy.deepcopy(nset)
        for other_layer in list(slide_nset.layers()):
            if other_layer.attrib['id'] == layer.attrib['id']:
                layer = other_layer
                continue
            if master_attr not in other_layer.attrib:
                set_style(other_layer, 'style', 'display', 'none')
#         for elem in other_layer.iter():
#           set_style(elem, 'style', 'display', 'none')
#         slide_nset.root.remove(other_layer)

        viewbox = None
        for j, ordr in enumerate(sorted(order)):
            if j not in selected_effects:
                continue
            outname = f'slide_{layer_fmt.format(i)}_effect_{effect_format.format(j)}.svg'
            for child in layer.iter():
                if slide_nset.display_by_effect(child, ordr) == JI_EFFECT_VIEW:
                    viewbox = child

            if pargs.simplify or viewbox:
                output_nset = copy.deepcopy(slide_nset)
                if pargs.simplify:
                    output_nset.remove_invisible(elems=itertools.chain.from_iterable(
                        lyr.iter() for lyr in output_nset.layers() if not is_hidden(lyr)
                    ))
            else:
                output_nset = slide_nset

            if viewbox is not None:
                viewbox_transform = list(output_nset.collect_transformations(viewbox))
                viewbox_transform = matrix_invert(viewbox_transform)
                try:
                    viewbox_value = (
                        f"{viewbox.attrib['x']} "
                        f"{viewbox.attrib['y']} "
                        f"{viewbox.attrib['width']} "
                        f"{viewbox.attrib['height']}"
                    )
                except KeyError as err:
                    LOGGER.error('Failed to set JessyInk view (missing attribute: %s).', err)
                else:
                    output_nset.root.attrib['viewBox'] = viewbox_value
                    # x<Min|Mid|Max>Y<Min|Mid|Max> <meet|slide|none>
                    output_nset.root.attrib['preserveAspectRatio'] = 'xMidYMid meet'
                    for some_layer in output_nset.layers():
                        layer_transform = parse_transform(some_layer)
                        transform_matrix = ', '.join(
                            str(x)
                            for x in matrix_multiply(viewbox_transform, layer_transform)
                        )
                        some_layer.attrib['transform'] = f'matrix({transform_matrix})'

            outpath = os.path.join(outdir, outname)
            output_nset.save(outpath)


# ----------------------------------- main ----------------------------------- #
def main(args=None):
    '''
    Parse command-line arguments and run the requested operations.
    '''
    parser = get_parser()
    pargs = parser.parse_args(args)
    if pargs.debug:
        log_level = logging.DEBUG
    elif pargs.verbose:
        log_level = logging.INFO
    else:
        log_level = logging.WARNING

    logging.basicConfig(
        format='%(levelname)s: %(message)s',
        level=log_level
    )
    try:
        func = pargs.func
    except AttributeError as err:
        raise GenericException(f'unrecognized command: {pargs.command}') from err
    func(pargs)


def run_main(args=None):
    '''
    Wrapper around main with basic exception handling.
    '''
    try:
        main(args=args)
    except (KeyboardInterrupt, BrokenPipeError):
        pass
    except (OSError, GenericException, ET.ParseError) as err:
        sys.exit(f'error: {err}')


if __name__ == '__main__':
    run_main()
